/*
 * livepatch_bsc1204486
 *
 * Fix for CVE-2022-3577, bsc#1204486
 *
 *  Upstream commit:
 *  fc4ef9d57249 ("HID: bigben: fix slab-out-of-bounds Write in bigben_probe")
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  not affected
 *
 *  SLE15-SP2 commit:
 *  e57339bf6fbb381b643ddfcacbbe0b07eedd1984
 *
 *  SLE15-SP3 commit:
 *  eb18a6409032033d480a7ad3d60105cdd0ca47cf
 *
 *  SLE15-SP4 commit:
 *  4e6de4ed4c95a6d4791c7bd841cbc590302b6b4b
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_HID_BIGBEN_FF)

#if !IS_MODULE(CONFIG_HID_BIGBEN_FF)
#error "Live patch supports only CONFIG_HID_BIGBEN_FF=m"
#endif

/* klp-ccp: from drivers/hid/hid-bigbenff.c */
#include <linux/input.h>

/* klp-ccp: from include/linux/input.h */
static int (*klpe_input_ff_create_memless)(struct input_dev *dev, void *data,
		int (*play_effect)(struct input_dev *, void *, struct ff_effect *));

/* klp-ccp: from drivers/hid/hid-bigbenff.c */
#include <linux/leds.h>
#include <linux/hid.h>

#define NUM_LEDS 4

struct bigben_device {
	struct hid_device *hid;
	struct hid_report *report;
	bool removed;
	u8 led_state;         /* LED1 = 1 .. LED4 = 8 */
	u8 right_motor_on;    /* right motor off/on 0/1 */
	u8 left_motor_force;  /* left motor force 0-255 */
	struct led_classdev *leds[NUM_LEDS];
	bool work_led;
	bool work_ff;
	struct work_struct worker;
};

static void (*klpe_bigben_worker)(struct work_struct *work);

static int (*klpe_hid_bigben_play_effect)(struct input_dev *dev, void *data,
			 struct ff_effect *effect);

static void (*klpe_bigben_set_led)(struct led_classdev *led,
	enum led_brightness value);

static enum led_brightness (*klpe_bigben_get_led)(struct led_classdev *led);

int klpp_bigben_probe(struct hid_device *hid,
	const struct hid_device_id *id)
{
	struct bigben_device *bigben;
	struct hid_input *hidinput;
	struct list_head *report_list;
	struct led_classdev *led;
	char *name;
	size_t name_sz;
	int n, error;

	bigben = devm_kzalloc(&hid->dev, sizeof(*bigben), GFP_KERNEL);
	if (!bigben)
		return -ENOMEM;
	hid_set_drvdata(hid, bigben);
	bigben->hid = hid;
	bigben->removed = false;

	error = hid_parse(hid);
	if (error) {
		hid_err(hid, "parse failed\n");
		return error;
	}

	error = hid_hw_start(hid, HID_CONNECT_DEFAULT & ~HID_CONNECT_FF);
	if (error) {
		hid_err(hid, "hw start failed\n");
		return error;
	}

	report_list = &hid->report_enum[HID_OUTPUT_REPORT].report_list;
	bigben->report = list_entry(report_list->next,
		struct hid_report, list);

	/*
	 * Fix CVE-2022-3577
	 *  +5 lines
	 */
	if (list_empty(&hid->inputs)) {
		hid_err(hid, "no inputs found\n");
		error = -ENODEV;
		goto error_hw_stop;
	}

	hidinput = list_first_entry(&hid->inputs, struct hid_input, list);
	set_bit(FF_RUMBLE, hidinput->input->ffbit);

	INIT_WORK(&bigben->worker, (*klpe_bigben_worker));

	error = (*klpe_input_ff_create_memless)(hidinput->input, NULL,
		(*klpe_hid_bigben_play_effect));
	if (error)
		goto error_hw_stop;

	name_sz = strlen(dev_name(&hid->dev)) + strlen(":red:bigben#") + 1;

	for (n = 0; n < NUM_LEDS; n++) {
		led = devm_kzalloc(
			&hid->dev,
			sizeof(struct led_classdev) + name_sz,
			GFP_KERNEL
		);
		if (!led) {
			error = -ENOMEM;
			goto error_hw_stop;
		}
		name = (void *)(&led[1]);
		snprintf(name, name_sz,
			"%s:red:bigben%d",
			dev_name(&hid->dev), n + 1
		);
		led->name = name;
		led->brightness = (n == 0) ? LED_ON : LED_OFF;
		led->max_brightness = 1;
		led->brightness_get = (*klpe_bigben_get_led);
		led->brightness_set = (*klpe_bigben_set_led);
		bigben->leds[n] = led;
		error = devm_led_classdev_register(&hid->dev, led);
		if (error)
			goto error_hw_stop;
	}

	/* initial state: LED1 is on, no rumble effect */
	bigben->led_state = BIT(0);
	bigben->right_motor_on = 0;
	bigben->left_motor_force = 0;
	bigben->work_led = true;
	bigben->work_ff = true;
	schedule_work(&bigben->worker);

	hid_info(hid, "LED and force feedback support for BigBen gamepad\n");

	return 0;

error_hw_stop:
	hid_hw_stop(hid);
	return error;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1204486.h"
#include "../kallsyms_relocs.h"

#define LIVEPATCHED_MODULE "hid_bigbenff"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "input_ff_create_memless", (void *)&klpe_input_ff_create_memless,
	  "ff_memless" },
	{ "bigben_get_led", (void *)&klpe_bigben_get_led, "hid_bigbenff" },
	{ "bigben_set_led", (void *)&klpe_bigben_set_led, "hid_bigbenff" },
	{ "bigben_worker", (void *)&klpe_bigben_worker, "hid_bigbenff" },
	{ "hid_bigben_play_effect", (void *)&klpe_hid_bigben_play_effect,
	  "hid_bigbenff" },
};

static int livepatch_bsc1204486_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LIVEPATCHED_MODULE))
		return 0;

	ret = klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
	WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block livepatch_bsc1204486_module_nb = {
	.notifier_call = livepatch_bsc1204486_module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1204486_init(void)
{
	int ret;
	struct module *mod;

	ret = klp_kallsyms_relocs_init();
	if (ret)
		return ret;

	ret = register_module_notifier(&livepatch_bsc1204486_module_nb);
	if (ret)
		return ret;

	rcu_read_lock_sched();
	mod = (*klpe_find_module)(LIVEPATCHED_MODULE);
	if (!try_module_get(mod))
		mod = NULL;
	rcu_read_unlock_sched();

	if (mod) {
		ret = klp_resolve_kallsyms_relocs(klp_funcs,
						  ARRAY_SIZE(klp_funcs));
	}

	if (ret)
		unregister_module_notifier(&livepatch_bsc1204486_module_nb);

	module_put(mod);
	return ret;
}

void livepatch_bsc1204486_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1204486_module_nb);
}

#endif /* IS_ENABLED(CONFIG_HID_BIGBEN_FF) */
